@@ -118,6 +118,18 @@ def wrapCommandArguments(cmd, arguments):
     ret.append(line)
     return ret
 
+def getGcNum(options, root, key):
+    key = "jobs.gc." + ("root" if root else "deps") + "." + key
+    try:
+        val = options.get(key, "1" if key == "jobs.gc.deps.artifacts" else "0")
+        num = int(val)
+    except ValueError:
+        raise ParseError("Invalid option '{}': '{}'".format(key, val))
+    if num <= 0:
+        return "-1"
+    else:
+        return str(num)
+
 class SpecHasher:
     """Track digest calculation and output as spec for bob-hash-engine.
 
@@ -159,6 +171,16 @@ def genUuid():
     ret = "".join(random.sample("0123456789abcdef", 8))
     return ret[:4] + '-' + ret[4:]
 
+def getAuditMeta(options):
+    return {
+        k[len("audit.meta."):] : v for k, v in sorted(options.items())
+        if k.startswith("audit.meta.")
+    }
+
+def verifyAuditMeta(meta):
+    valid = re.compile(r"[0-9A-Za-z._-]+")
+    return all(valid.fullmatch(k) for k in meta.keys())
+
 class JenkinsJob:
     def __init__(self, name, displayName, nameCalculator, recipe, archiveBackend):
         self.__name = name
@@ -297,8 +319,7 @@ class JenkinsJob:
             cmds.append("fi")
             cmds.append("")
 
-        if d.getJenkinsScript() is not None:
-            cmds.append("bob _invoke {} -vv".format(self._specName(d)))
+        cmds.append("bob _invoke {} -vv".format(self._specName(d)))
 
         return "\n".join(cmds)
 
@@ -326,13 +347,17 @@ class JenkinsJob:
             fingerprint = self._fingerprintName(step)
             if fingerprint: fingerprint = "{sha1\n<" + fingerprint + "\n}"
             ret.extend(step.getDigest(lambda s: JenkinsJob._buildIdName(s), True,
-                SpecHasher, fingerprint=fingerprint, platform='p'))
+                SpecHasher, fingerprint=fingerprint, platform='p',
+                relaxTools=True))
 
         ret.append("EOF")
         return ret
 
-    def dumpStepAuditGen(self, step):
-        cmd = [
+    def dumpStepAuditGen(self, step, auditMeta):
+        cmd = []
+        for k, v in sorted(auditMeta.items()):
+            cmd += ["-D", quote(k), quote(v)]
+        cmd += [
             "-D", "bob", BOB_VERSION,
             "-D", "recipe", step.getPackage().getRecipe().getName(),
             "-D", "package", "/".join(step.getPackage().getStack()),
@@ -348,9 +373,7 @@ class JenkinsJob:
         if recipesAudit is not None:
             cmd.extend(["--recipes",
                 quote(json.dumps(recipesAudit.dump(), sort_keys=True))])
-        # environment is only persisted if a shell script is run
-        if step.getJenkinsScript() is not None:
-            cmd.extend(["--env", JenkinsJob._envName(step)])
+        cmd.extend(["--env", JenkinsJob._envName(step)])
         if step.isCheckoutStep():
             for scm in step.getScmList():
                 auditSpec = scm.getAuditSpec()
@@ -586,24 +609,8 @@ class JenkinsJob:
                 xml.etree.ElementTree.SubElement(
                     root, "displayName").text = self.__displayName
             xml.etree.ElementTree.SubElement(root, "keepDependencies").text = "false"
-            properties = xml.etree.ElementTree.SubElement(root, "properties")
-            if not self.__isRoot:
-                # only retain one artifact per non-root job
-                discard = xml.etree.ElementTree.fromstring("""
-                    <jenkins.model.BuildDiscarderProperty>
-                      <strategy class="hudson.tasks.LogRotator">
-                        <daysToKeep>-1</daysToKeep>
-                        <numToKeep>-1</numToKeep>
-                        <artifactDaysToKeep>-1</artifactDaysToKeep>
-                        <artifactNumToKeep>1</artifactNumToKeep>
-                      </strategy>
-                    </jenkins.model.BuildDiscarderProperty>""")
-                properties.append(discard)
-            if (nodes != ''):
-                xml.etree.ElementTree.SubElement(root, "assignedNode").text = nodes
-                xml.etree.ElementTree.SubElement(root, "canRoam").text = "false"
-            else:
-                xml.etree.ElementTree.SubElement(root, "canRoam").text = "true"
+            xml.etree.ElementTree.SubElement(root, "properties")
+            xml.etree.ElementTree.SubElement(root, "canRoam").text = "true"
             xml.etree.ElementTree.SubElement(root, "disabled").text = "false"
             xml.etree.ElementTree.SubElement(
                 root, "blockBuildWhenDownstreamBuilding").text = "false"
@@ -630,6 +637,36 @@ class JenkinsJob:
         xml.etree.ElementTree.SubElement(
             scmTrigger, "ignorePostCommitHooks").text = ("true" if isTrue(options.get("scm.ignore-hooks", "false")) else "false")
 
+        numToKeep = getGcNum(options, self.__isRoot, "builds")
+        artifactNumToKeep = getGcNum(options, self.__isRoot, "artifacts")
+        discard = root.find("./properties/jenkins.model.BuildDiscarderProperty/strategy[@class='hudson.tasks.LogRotator']")
+        if numToKeep != "-1" or artifactNumToKeep != "-1":
+            if discard is None:
+                discard = xml.etree.ElementTree.fromstring("""
+                      <strategy class="hudson.tasks.LogRotator">
+                        <daysToKeep>-1</daysToKeep>
+                        <numToKeep>-1</numToKeep>
+                        <artifactDaysToKeep>-1</artifactDaysToKeep>
+                        <artifactNumToKeep>1</artifactNumToKeep>
+                      </strategy>""")
+                xml.etree.ElementTree.SubElement(
+                    root.find("properties"),
+                    "jenkins.model.BuildDiscarderProperty").append(discard)
+            discard.find("numToKeep").text = numToKeep
+            discard.find("artifactNumToKeep").text = artifactNumToKeep
+        elif discard is not None:
+            properties = root.find("properties")
+            properties.remove(properties.find("jenkins.model.BuildDiscarderProperty"))
+
+        if (nodes != ''):
+            assignedNode = root.find("assignedNode")
+            if assignedNode is None:
+                assignedNode = xml.etree.ElementTree.SubElement(root, "assignedNode")
+            assignedNode.text = nodes
+            root.find("canRoam").text = "false"
+        else:
+            root.find("canRoam").text = "true"
+
         sharedDir = options.get("shared.dir", "${JENKINS_HOME}/bob")
 
         specCmds = []
@@ -808,11 +845,10 @@ class JenkinsJob:
         # checkout steps
         checkoutSCMs = []
         for d in sorted(self.__checkoutSteps.values()):
-            if d.getJenkinsScript() is not None:
-                checkout = xml.etree.ElementTree.SubElement(
-                    builders, "hudson.tasks.Shell")
-                xml.etree.ElementTree.SubElement(
-                    checkout, "command").text = self.dumpStep(d, windows, [])
+            checkout = xml.etree.ElementTree.SubElement(
+                builders, "hudson.tasks.Shell")
+            xml.etree.ElementTree.SubElement(
+                checkout, "command").text = self.dumpStep(d, windows, [])
             checkoutSCMs.extend(d.getJenkinsXml(credentials, options))
 
         if len(checkoutSCMs) > 1:
@@ -855,6 +891,9 @@ class JenkinsJob:
         xml.etree.ElementTree.SubElement(
             checkout, "command").text = "\n".join(buildIdCalc)
 
+        # extra audit trail meta variables
+        auditMeta = getAuditMeta(options)
+
         # generate audit trail of checkout steps
         if self.__checkoutSteps:
             checkoutAudit = [
@@ -862,7 +901,7 @@ class JenkinsJob:
                 "# generate audit trail of checkout step(s)"
             ]
             for d in sorted(self.__checkoutSteps.values()):
-                checkoutAudit.append(self.dumpStepAuditGen(d))
+                checkoutAudit.append(self.dumpStepAuditGen(d, auditMeta))
             audit = xml.etree.ElementTree.SubElement(
                 builders, "hudson.tasks.Shell")
             xml.etree.ElementTree.SubElement(
@@ -900,7 +939,7 @@ class JenkinsJob:
                     self.dumpStep(d, windows, affectedPackageSteps),
                     "", "# generate audit trail",
                     "cd \"$WORKSPACE\"",
-                    self.dumpStepAuditGen(d)
+                    self.dumpStepAuditGen(d, auditMeta)
                 ])
 
         # package steps
@@ -912,7 +951,7 @@ class JenkinsJob:
                 self.dumpStep(d, windows, [d]),
                 "", "# generate audit trail",
                 "cd \"$WORKSPACE\"",
-                self.dumpStepAuditGen(d),
+                self.dumpStepAuditGen(d, auditMeta),
                 "", "# pack result for archive and inter-job exchange",
                 "cd \"$WORKSPACE\"",
                 "tar zcfv {TGZ} -H pax --pax-option=\"bob-archive-vsn=1\" --transform='s|^{AUDIT}|meta/audit.json.gz|' --transform='s|^{WSP_PATH}|content|' {AUDIT} {WSP_PATH}".format(
@@ -1244,21 +1283,21 @@ def jenkinsNamePersister(jenkins, wrapFmt, uuid):
         if uuid: ret = ret + "-" + uuid
         return ret
 
-    def fmt(step, mode, props, referrer):
+    def fmt(step, mode, props):
         if mode == 'workspace':
             return persist(step, props)
         else:
-            assert mode == 'exec'
-            if referrer.getSandbox() is None:
-                return persist(step, props)
-            else:
-                return os.path.join("/bob", asHexStr(step.getVariantId()), "workspace")
+            assert mode == 'storage'
+            return step.getWorkspacePath()
 
     return fmt
 
 def genJenkinsJobs(recipes, jenkins):
     jobs = {}
     config = BobState().getJenkinsConfig(jenkins)
+    recipes.parse(config.get('defines', {}),
+                  "msys" if config.get("windows", False) else "linux")
+
     prefix = config["prefix"]
     options = config.get("options", {})
     archiveHandler = getArchiver(recipes)
@@ -1269,12 +1308,10 @@ def genJenkinsJobs(recipes, jenkins):
         if not archiveHandler.canUploadJenkins() or not archiveHandler.canDownloadJenkins():
             raise ParseError("No archive for up and download found but artifacts.copy using archive enabled!")
     nameFormatter = recipes.getHook('jenkinsNameFormatter')
-    windows = config.get("windows", False)
+
     packages = recipes.generatePackages(
         jenkinsNamePersister(jenkins, nameFormatter, config.get('uuid')),
-        config.get('defines', {}),
-        config.get('sandbox', False),
-        "msys" if windows else "linux")
+        config.get('sandbox', False))
     nameCalculator = JobNameCalculator(prefix)
     rootPackages = []
     for r in config["roots"]: rootPackages.extend(packages.queryPackagePath(r))
@@ -1360,6 +1397,8 @@ def doJenkinsAdd(recipes, argv):
             parser.error("Archive sharing can not be used without upload enabled! Exiting..", file=sys.stderr)
         if not args.download:
             parser.error("Archive sharing can not be used without download enabled! Exiting..", file=sys.stderr)
+    if not verifyAuditMeta(getAuditMeta(options)):
+        parser.error("Invalid audit meta variable name")
 
     if args.name in BobState().getAllJenkins():
         print("Jenkins '{}' already added.".format(args.name), file=sys.stderr)
@@ -1410,8 +1449,6 @@ def doJenkinsExport(recipes, argv):
               file=sys.stderr)
         sys.exit(1)
 
-    jenkinsJobCreate = recipes.getHookStack('jenkinsJobCreate')
-
     jobs = genJenkinsJobs(recipes, args.name)
     buildOrder = genJenkinsBuildOrder(jobs)
     config = BobState().getJenkinsConfig(args.name)
@@ -1422,6 +1459,7 @@ def doJenkinsExport(recipes, argv):
     options = config.get("options", {})
     authtoken = config.get("authtoken")
 
+    jenkinsJobCreate = recipes.getHookStack('jenkinsJobCreate')
     for j in buildOrder:
         job = jobs[j]
         info = {
@@ -2155,6 +2193,8 @@ def doJenkinsSetOptions(recipes, argv):
             parser.error("Archive sharing can not be used without upload enabled! Exiting..")
         if not config.get('download', False):
             parser.error("Archive sharing can not be used without download enabled! Exiting..")
+    if not verifyAuditMeta(getAuditMeta(options)):
+        parser.error("Invalid audit meta variable name")
 
     BobState().setJenkinsConfig(args.name, config)
 
@@ -2190,7 +2230,6 @@ def doJenkins(argv, bobRoot):
     recipes = RecipeSet()
     recipes.defineHook('jenkinsNameFormatter', jenkinsNameFormatter)
     recipes.setConfigFiles(args.configFile)
-    recipes.parse()
 
     if args.subcommand in availableJenkinsCmds:
         BobState().setAsynchronous()
