@@ -34,8 +34,33 @@ bob-libc-version()
     # Try glibc first
     cat >conftest.c <<EOF
 #include <stdio.h>
+#if defined(__MSYS__) || defined(__CYGWIN__)
+#include <cygwin/version.h>
+#elif defined(__MINGW32__) || defined(__MINGW64__)
+#include <windows.h>
+#else
 #include <gnu/libc-version.h>
-int main(){ printf("glibc %s\n", gnu_get_libc_version()); return 0; }
+#endif
+int main(void)
+{
+#if defined(CYGWIN_VERSION_DLL_IDENTIFIER)
+    printf("%s/%d.%d.%d\n", CYGWIN_VERSION_DLL_IDENTIFIER,
+        CYGWIN_VERSION_DLL_EPOCH,
+        CYGWIN_VERSION_DLL_MAJOR,
+        CYGWIN_VERSION_DLL_MINOR);
+#elif defined(__MSVCRT_VERSION__)
+    printf("mingw%d/msvcrt-%03x\n",
+#ifdef __MINGW64__
+        64,
+#else
+        32,
+#endif
+        __MSVCRT_VERSION__);
+#else
+    printf("glibc %s\n", gnu_get_libc_version());
+#endif
+    return 0;
+}
 EOF
     if ${1:-${CC:-cc}} -o conftest conftest.c >/dev/null ; then
         ./conftest && return 0
@@ -223,7 +248,7 @@ class BashLanguage:
             return p
 
     @staticmethod
-    def __formatSetup(spec, keepEnv):
+    def __formatProlog(spec, keepEnv):
         env = { key: quote(value) for (key, value) in spec.env.items() }
         env.update({
             "PATH": ":".join(
@@ -267,6 +292,15 @@ class BashLanguage:
         ])
         return "\n".join(ret)
 
+    @staticmethod
+    def __formatSetup(spec):
+        return "\n".join([
+            "",
+            "# Recipe setup script",
+            spec.setupScript,
+            "cd \"${BOB_CWD}\"",
+        ])
+
     @staticmethod
     def __formatScript(spec):
         colorize = not spec.isJenkins
@@ -275,7 +309,7 @@ class BashLanguage:
         else:
             envFile = None
         ret = [
-            BashLanguage.__formatSetup(spec, False),
+            BashLanguage.__formatProlog(spec, False),
             "",
             "# Setup",
             "declare -p > {}".format(quote(BashLanguage.__munge(envFile))) if envFile else "",
@@ -306,8 +340,10 @@ class BashLanguage:
                 trap 'for i in "${_BOB_TMP_CLEANUP[@]-}" ; do /bin/rm -f "$i" ; done' EXIT
                 set -o errtrace -o nounset -o pipefail
                 """),
-            "# Recipe script",
-            spec.script,
+            BashLanguage.__formatSetup(spec),
+            "",
+            "# Recipe main script",
+            spec.mainScript,
         ]
         return "\n".join(ret)
 
@@ -325,7 +361,8 @@ class BashLanguage:
     def setupShell(spec, tmpDir, keepEnv):
         realScriptFile, execScriptFile = BashLanguage.__scriptFilePaths(spec, tmpDir)
         with open(realScriptFile, "w") as f:
-            f.write(BashLanguage.__formatSetup(spec, keepEnv))
+            f.write(BashLanguage.__formatProlog(spec, keepEnv))
+            f.write(BashLanguage.__formatSetup(spec))
 
         args = ["bash", "--rcfile", BashLanguage.__munge(execScriptFile), "-s", "--"]
         args.extend(BashLanguage.__munge(os.path.abspath(a)) for a in spec.args)
@@ -410,7 +447,7 @@ class PwshLanguage:
         """)
 
     @staticmethod
-    def __formatSetup(spec):
+    def __formatProlog(spec):
         pathSep = ";" if sys.platform == "win32" else ":"
         env = { key: escapePwsh(value) for (key, value) in spec.env.items() }
         env.update({
@@ -445,6 +482,15 @@ class PwshLanguage:
         ]
         return "\n".join(ret)
 
+    @staticmethod
+    def __formatSetup(spec):
+        return "\n".join([
+            "",
+            "# Recipe setup script",
+            spec.setupScript,
+            "cd $Env:BOB_CWD",
+        ])
+
     @staticmethod
     def __formatScript(spec, trace):
         if spec.envFile:
@@ -452,7 +498,7 @@ class PwshLanguage:
         else:
             envFile = None
         ret = [
-            PwshLanguage.__formatSetup(spec),
+            PwshLanguage.__formatProlog(spec),
             "",
             "# Setup",
             dedent("""\
@@ -480,9 +526,10 @@ class PwshLanguage:
                 try {
                     $_BOB_TMP_CLEANUP = @()
                 """),
-            "# BEGIN BUILD SCRIPT",
-            spec.script,
-            "# END BUILD SCRIPT",
+            PwshLanguage.__formatSetup(spec),
+            "",
+            "# Recipe main script",
+            spec.mainScript,
             dedent("""\
                 } finally {
                     foreach($f in $_BOB_TMP_CLEANUP) {
@@ -506,6 +553,7 @@ class PwshLanguage:
     def setupShell(spec, tmpDir, keepEnv):
         realScriptFile, execScriptFile = PwshLanguage.__scriptFilePaths(spec, tmpDir)
         with open(realScriptFile, "w") as f:
+            f.write(PwshLanguage.__formatProlog(spec))
             f.write(PwshLanguage.__formatSetup(spec))
 
         interpreter = "powershell" if isWindows() else "pwsh"
@@ -577,7 +625,7 @@ class StepSpec:
             'env' : dict(step.getEnv()),
             'paths' : step.getPaths(),
             'libraryPaths' : step.getLibraryPaths(),
-            'workspace' : (step.getWorkspacePath(), step.getExecPath()),
+            'workspace' : (step.getStoragePath(), step.getExecPath()),
             'args' : [ a.getExecPath(step) for a in step.getArguments() ],
             'allPaths' : sorted([
                 (a.getPackage().getName(), a.getExecPath(step))
@@ -603,12 +651,12 @@ class StepSpec:
         # fetch sandbox if configured
         if step.getSandbox() is not None:
             d['sandbox'] = s = {
-                'root' : step.getSandbox().getStep().getWorkspacePath(),
+                'root' : step.getSandbox().getStep().getStoragePath(),
                 'paths' : step.getSandbox().getPaths(),
                 'hostMounts' : step.getSandbox().getMounts(),
                 'netAccess' : step.hasNetAccess(),
                 'depMounts' : [
-                    (dep.getWorkspacePath(), dep.getExecPath(step))
+                    (dep.getStoragePath(), dep.getExecPath(step))
                     for dep in step.getAllDepSteps() if dep.isValid()
                 ],
             }
@@ -618,11 +666,12 @@ class StepSpec:
             while extra.isValid() and len(extra.getArguments()) > 0:
                 extra = extra.getArguments()[0]
                 if extra.isValid():
-                    s['depMounts'].append((extra.getWorkspacePath(), extra.getExecPath(step)))
+                    s['depMounts'].append((extra.getStoragePath(), extra.getExecPath(step)))
 
         d['preRunCmds'] = step.getJenkinsPreRunCmds() if isJenkins else step.getPreRunCmds()
-        d['script'] = step.getJenkinsScript() if isJenkins else step.getScript()
-        d['postRunCmds'] = step.getJenkinsPostRunCmds() if isJenkins else step.getPostRunCmds()
+        d['setupScript'] = step.getSetupScript()
+        d['mainScript'] = step.getMainScript()
+        d['postRunCmds'] = step.getPostRunCmds()
         d['fingerprintScript'] = step._getFingerprintScript()
 
         return self
@@ -716,8 +765,12 @@ class StepSpec:
         return self.__data['preRunCmds']
 
     @property
-    def script(self):
-        return self.__data['script']
+    def setupScript(self):
+        return self.__data['setupScript']
+
+    @property
+    def mainScript(self):
+        return self.__data['mainScript']
 
     @property
     def fingerprintScript(self):
